package shell

import (
	"errors"
	"fmt"
	"os"
	"time"

	"github.com/go-cmd/cmd"
	"go.wit.com/log"
)

// this is a simplified interaction with the excellent
// go-cmd/cmd package to work 'shell' like.

// in all cases here, STDERR -> STDOUT
// If you want the output from whatever you run
// to be captured like it appears when you see it
// on the command line, this is what this tries to do

/*
	if r := shell.Run([]{"ping", "-c", "3", "localhost"}); r.Error == nil {
		if r.Exit == 0 {
			log.Println("ran ok")
		} else {
			log.Println("ran")
		}
		// all stdout/stderr captured in r.Stdout
	}
*/

// shortcut, sends a blank value for pwd
// which means the exec Dir is not set
// echos output (otherwise use RunQuiet)
func Run(argv []string) cmd.Status {
	return PathRun("", argv)
}

// exec the cmd at a filepath. this does not change the working directory
// sets the exec dir if it's not an empty string
// combines stdout and stderr
// echo's output (otherwise use PathRunQuiet()
// this is basically the exact example from the go-cmd/cmd devs
// where the have rocked out a proper smart read on both filehandles
// https://dave.cheney.net/2013/04/30/curious-channels
func PathRun(path string, argv []string) cmd.Status {
	return PathRunLog(path, argv, INFO)
}

// the actual wrapper around go-cmd/cmd
// adds a log Flag so that echo to stdout can be enabled/disabled
func PathRunLog(path string, argv []string, logf *log.LogFlag) cmd.Status {
	var save []string // combined stdout & stderr
	var arg0 string
	var args []string
	if logf == nil {
		logf = NOW
	}
	log.Log(logf, "shell.PathRunLog() Path =", path, "cmd =", argv)
	// Check if the slice has at least one element (the command name)
	if len(argv) == 0 {
		var s cmd.Status
		s.Error = errors.New("Error: Command slice is empty.")
		return s
	}
	if len(argv) == 1 {
		// Pass the first element as the command, and the rest as variadic arguments
		arg0 = argv[0]
	} else {
		arg0 = argv[0]
		args = argv[1:]
	}

	// Disable output buffering, enable streaming
	cmdOptions := cmd.Options{
		Buffered:  false,
		Streaming: true,
	}

	// Create Cmd with options
	envCmd := cmd.NewCmdOptions(cmdOptions, arg0, args...)
	if path != "" {
		// set the path for exec
		envCmd.Dir = path
	}

	// Print STDOUT and STDERR lines streaming from Cmd
	doneChan := make(chan struct{})
	go func() {
		defer close(doneChan)
		// Done when both channels have been closed
		// https://dave.cheney.net/2013/04/30/curious-channels
		for envCmd.Stdout != nil || envCmd.Stderr != nil {
			select {
			case line, open := <-envCmd.Stdout:
				if !open {
					envCmd.Stdout = nil
					continue
				}
				save = append(save, line)
				log.Log(logf, line)
				// fmt.Println(line)
			case line, open := <-envCmd.Stderr:
				if !open {
					envCmd.Stderr = nil
					continue
				}
				save = append(save, line)
				log.Log(logf, line)
				// fmt.Println(line)
			}
		}
	}()

	// Run and wait for Cmd to return, discard Status
	<-envCmd.Start()

	// Wait for goroutine to print everything
	<-doneChan

	s := envCmd.Status()
	s.Stdout = save
	return s
}

// uses the 'log' package to disable echo to STDOUT
// only echos if you enable the shell.INFO log flag
func RunQuiet(args []string) cmd.Status {
	return PathRunLog("", args, INFO)
}

// uses the 'log' package to disable echo to STDOUT
// only echos if you enable the shell.INFO log flag
func PathRunQuiet(pwd string, args []string) cmd.Status {
	return PathRunLog(pwd, args, INFO)
}

// send blank path == use current golang working directory
func RunRealtime(args []string) cmd.Status {
	return PathRunRealtime("", args)
}

func RunEcho(cmd []string) cmd.Status {
	result := RunQuiet(cmd)
	pwd, _ := os.Getwd()
	log.Warn("shell.RunEcho() cmd:", cmd, pwd)
	log.Warn("shell.RunEcho() Exit:", result.Exit)
	log.Warn("shell.RunEcho() Error:", result.Error)
	for _, line := range result.Stdout {
		log.Warn("STDOUT:", line)
	}
	for _, line := range result.Stderr {
		log.Warn("STDERR:", line)
	}
	return result
}

// echos twice a second if anything sends to STDOUT or STDERR
// not great, but it's really just for watching things run in real time anyway
// TODO: fix \r handling for things like git-clone so the terminal doesn't
// have to do a \n newline each time.
// TODO: add timeouts and status of things hanging around forever
func PathRunRealtime(pwd string, args []string) cmd.Status {
	// Check if the slice has at least one element (the command name)
	if len(args) == 0 {
		var s cmd.Status
		s.Error = errors.New("Error: Command slice is empty.")
		return s
	}

	// Start a long-running process, capture stdout and stderr
	a, b := RemoveFirstElement(args)
	findCmd := cmd.NewCmd(a, b...)
	if pwd != "" {
		findCmd.Dir = pwd
	}
	statusChan := findCmd.Start() // non-blocking

	ticker := time.NewTicker(100 * time.Microsecond)

	// this is interesting, maybe useful, but wierd, but neat. interesting even
	// Print last line of stdout every 2s
	go func() {
		// loop very quickly, but only print the line if it changes
		var lastout string
		var lasterr string
		for range ticker.C {
			status := findCmd.Status()
			n := len(status.Stdout)
			if n != 0 {
				newline := status.Stdout[n-1]
				if lastout != newline {
					lastout = newline
					log.Info(lastout)
				}
			}
			n = len(status.Stderr)
			if n != 0 {
				newline := status.Stderr[n-1]
				if lasterr != newline {
					lasterr = newline
					log.Info(lasterr)
				}
			}
			if status.Complete {
				return
			}
		}
	}()

	// Stop command after 1 hour
	go func() {
		<-time.After(1 * time.Hour)
		findCmd.Stop()
	}()

	// Check if command is done
	select {
	case finalStatus := <-statusChan:
		log.Info("finalStatus =", finalStatus.Exit, finalStatus.Error)
		return finalStatus
		// done
	default:
		// no, still running
	}

	// Block waiting for command to exit, be stopped, or be killed
	// there are things being left around here. debug this
	finalStatus := <-statusChan
	if len(finalStatus.Cmd) != 0 {
		if string(finalStatus.Cmd) != "go" {
			log.Info("shell.Run() ok goroutine end?", finalStatus.Cmd, finalStatus.Exit)
		}
	}
	return findCmd.Status()
}

// this is stuff from a long time ago that there must be a replacement for
func RemoveFirstElement(slice []string) (string, []string) {
	if len(slice) == 0 {
		return "", slice // Return the original slice if it's empty
	}
	return slice[0], slice[1:] // Return the slice without the first element
}

func RunVerbose(cmd []string) (*cmd.Status, error) {
	pwd, _ := os.Getwd()
	log.Info("Running:", pwd, cmd)
	r, err := RunStrict(cmd)
	if err != nil {
		log.Info("Error", cmd, err)
	}
	for _, line := range r.Stdout {
		log.Info(line)
	}
	for _, line := range r.Stderr {
		log.Info(line)
	}
	return r, err
}

func RunVerboseOnError(cmd []string) (*cmd.Status, error) {
	r, err := RunStrict(cmd)
	if err == nil {
		return r, err
	}
	pwd, _ := os.Getwd()
	log.Info("Run Error:", pwd, cmd, err)
	for _, line := range r.Stdout {
		log.Info(line)
	}
	for _, line := range r.Stderr {
		log.Info(line)
	}
	return r, err
}

func RunStrict(cmd []string) (*cmd.Status, error) {
	pwd, _ := os.Getwd()
	result := PathRunQuiet(pwd, cmd)
	if result.Error != nil {
		log.Warn(pwd, cmd, "wow. golang is cool. an os.Error:", result.Error)
		return &result, result.Error
	}
	if result.Exit != 0 {
		// log.Warn(cmd, "failed with", result.Exit, repo.GetGoPath())
		return &result, errors.New(fmt.Sprint(cmd, "failed with", result.Exit))
	}
	return &result, nil
}
